from PIL import ImageTk
from PIL import ImageFilter
from PIL import Image as Img

from math import gcd, ceil
from os.path import isfile as osisfile

from goopylib.util import GraphicsError, resource_path

from goopylib.objects.GraphicsObject import GraphicsObject
from goopylib.constants import ALIGN_OPTIONS, IMAGE_INTERPOLATIONS


class Image(GraphicsObject):
    id_count = 0

    default_sampling = "bicubic"
    texture_path = "textures/"

    def __init__(self, p, filepath, align="center", cursor="arrow", layer=0, tag=None, bounds=None):

        if not isinstance(p, list):
            raise GraphicsError(f"\n\nGraphicsError: Image anchor (p) must be a list in the form [x, y], not {p}")
        if align not in ALIGN_OPTIONS:
            raise GraphicsError(f"\n\nGraphicsError: Image align must be one of {ALIGN_OPTIONS}, not {align}")

        if osisfile(resource_path(Image.texture_path + filepath)):
            self.texture = Image.texture_path + filepath
        elif osisfile(resource_path(filepath)):
            self.texture = filepath
        else:
            raise GraphicsError("\n\nGraphicsError: Image filepath is not valid. No such file or directory as "
                                f"{filepath}")

        self.texture = resource_path(self.texture)
        self.reprpath = f"\"{filepath}\""

        self.img_PIL = Img.open(self.texture, )
        self.img = ImageTk.PhotoImage(self.img_PIL)
        self.original_img = self.img_PIL.copy()
        self.transforming_img = self.img_PIL.copy()

        self.align = align
        self.anchor = p.copy()

        self.x = self.anchor[0]
        self.y = self.anchor[1]
        self.initial_width = self.get_width()
        self.initial_height = self.get_height()
        self.resized = False

        self.contrast = 0
        self.blur_amount = 0

        self.update_required = False

        GraphicsObject.__init__(self, [], cursor=cursor, layer=layer, tag=tag, bounds=bounds)

        self.image_id = Image.id_count
        Image.id_count = Image.id_count + 1

    # -------------------------------------------------------------------------
    # INTERNAL FUNCTIONS

    def __repr__(self):
        return f"Image({self.anchor}, {self.reprpath})"

    def _draw(self, canvas, options):
        x, y = self.anchor

        if self.align == "center":
            pass
        elif self.align == "bottom":
            y += self.initial_height / -2
        elif self.align == "top":
            y += self.initial_height / 2
        elif self.align == "left":
            x += self.initial_width / 2
        elif self.align == "right":
            x += self.initial_width / -2
        elif self.align == "bottomleft":
            y += self.initial_height / -2
            x += self.initial_width / 2
        elif self.align == "bottomright":
            y += self.initial_height / 2
            x += self.initial_width / -2
        elif self.align == "topleft":
            x += self.initial_width / 2
            y += self.initial_height / 2
        elif self.align == "topright":
            x += self.initial_width / -2
            y += self.initial_height / 2

        if canvas.trans.x_scale != 1 and canvas.trans.y_scale != 1:
            if not self.resized:
                init_width = abs(self.initial_width / canvas.trans.x_scale)
                init_height = abs(self.initial_height / canvas.trans.y_scale)
                self.resize(ceil(init_width), ceil(init_height), sampling=Image.default_sampling, _external_call=False)
                self.resized = True
            x, y = canvas.to_screen(x, y)

        if self.update_required:
            self.update()
            self.update_required = False

        return canvas.create_image(x, y, image=self.img)

    def _move(self, dx, dy):
        self.anchor[0] += dx
        self.anchor[1] += dy

    def _rotate(self, dr, sampling=None, center=None):
        if sampling is None:
            sampling = Image.default_sampling
        if center is not None:
            self.transforming_img = self.original_img.rotate(angle=-self.rotation, expand=True,
                                                             resample=IMAGE_INTERPOLATIONS[sampling],
                                                             center=(center.x, center.y))
        else:
            self.transforming_img = self.original_img.rotate(angle=-self.rotation, expand=True,
                                                             resample=IMAGE_INTERPOLATIONS[sampling])
        self.img_PIL = self.transforming_img.copy()
        self.update_required = True
        return self

    def update(self):
        try:
            self.graphwin.delete(self.id)
            self.img = ImageTk.PhotoImage(self.img_PIL, master=self.graphwin.master)
        except AttributeError:
            self.img = ImageTk.PhotoImage(self.img_PIL)

        self._update_layer()

    # -------------------------------------------------------------------------
    # OTHER & IMPORTANT FUNCTIONS

    def is_clicked(self, mouse_pos):
        if mouse_pos is None:
            return False
        else:
            if self.bounds is None:
                if self.drawn:
                    width, height = abs(self.img.width() * self.graphwin.trans.x_scale) // 2, \
                                    abs(self.img.height() * self.graphwin.trans.y_scale) // 2
                else:
                    width, height = self.img.width() // 2, self.img.height() // 2

                if (self.anchor[0] - width < mouse_pos[0] < self.anchor[0] + width) and (
                        self.anchor[1] - height < mouse_pos[1] < self.anchor[1] + height):
                    return True
                else:
                    return False
            else:
                return self.bounds.is_clicked(mouse_pos)

    def clone(self):
        other = Image(self.anchor.copy(), self.texture)
        other.config = self.config.copy()
        return other

    def save(self, filename):
        """Saves the pixmap image to filename.
        The format for the save image is determined from the filname extension.
        """

    # -------------------------------------------------------------------------
    # IMAGE COLOUR & ENHANCE FUNCTIONS

    def set_contrast(self, level):
        self.change_contrast(level - self.contrast)
        return self

    def change_contrast(self, level):
        if not (isinstance(level, int) or isinstance(level, float)):
            raise GraphicsError(f"\n\nGraphicsError: contrast level must be an integer or float, not {level}")

        self.contrast += level

        # Code from https://stackoverflow.com/questions/42045362/change-contrast-of-image-in-pil
        factor = (259 * (level + 255)) / (255 * (259 - level))

        def contrast(c):
            return 128 + factor * (c - 128)

        self.original_img = self.original_img.point(contrast)
        self.img_PIL = self.img_PIL.point(contrast)
        self.update_required = True

        self._update_layer()
        return self

    def reset_contrast(self):
        self.set_contrast(0)

    # -------------------------------------------------------------------------
    # IMAGE MANIPULATION FUNCTIONS

    def crop(self, left=0, top=0, right=500, down=500, align="center"):

        if not (isinstance(left, float) or isinstance(left, int)):
            raise GraphicsError(f"\n\nGraphicsError: The crop left argument must be an integer or float, not {left}")
        if not (isinstance(right, float) or isinstance(right, int)):
            raise GraphicsError(f"\n\nGraphicsError: The crop right argument must be an integer or float, not {right}")
        if not (isinstance(top, float) or isinstance(top, int)):
            raise GraphicsError(f"\n\nGraphicsError: The crop top argument must be an integer or float, not {top}")
        if not (isinstance(down, float) or isinstance(down, int)):
            raise GraphicsError(f"\n\nGraphicsError: The crop down argument must be an integer or float, not {down}")

        if align != "center":
            valid = False
            if "top" in align:
                self.move_y((self.get_height() - (down - top)) / -2)
                valid = True
            elif "bottom" in align:
                self.move_y((self.get_height() - (down - top)) / 2)
                valid = True

            if "left" in align:
                self.move_x((self.get_width() - (right - left)) / -2)
                valid = True
            elif "right" in align:
                self.move_x((self.get_width() - (right - left)) / 2)
                valid = True

            if not valid:
                raise GraphicsError(f"\n\nGraphicsError: Crop align must be one of {ALIGN_OPTIONS}, not {align}")

        self.img_PIL = self.img_PIL.crop(box=(left, top, right, down))
        self.transforming_img = self.transforming_img.crop(box=(left, top, right, down))

        self.update_required = True

        self._update_layer()
        return self

    # Blending & Compositing Functions

    def blend(self, img, alpha):
        if not isinstance(img, Image):
            raise GraphicsError("\n\nGraphicsError: img argument for blending must be another goopylib_b Image Object, "
                                f"not {img}")
        if not (isinstance(alpha, int) or isinstance(alpha, float)):
            raise GraphicsError(f"\n\nGraphicsError: Blend alpha value must be an integer or float, not {alpha}")
        if not (0 <= alpha <= 1):
            raise GraphicsError(f"\n\nGraphicsError: Blend alpha value must be between 0 and 1, not {alpha}")

        self.img_PIL = Img.blend(self.img_PIL, img.img_PIL, alpha)
        self.transforming_img = self.img_PIL
        self.update_required = True

        self._update_layer()
        return self

    def alpha_composite(self, img):
        if not isinstance(img, Image):
            raise GraphicsError("\n\nGraphicsError: img argument for composite must be another goopylib_b Image Object, "
                                f"not {img}")

        self.img_PIL = Img.alpha_composite(self.img_PIL, img.img_PIL)
        self.transforming_img = self.img_PIL
        self.update_required = True

        self._update_layer()
        return self

    def composite(self, img, img_mask):
        if not isinstance(img, Image):
            raise GraphicsError("\n\nGraphicsError: img argument for composite must be another goopylib_b Image Object, "
                                f"not {img}")
        if not isinstance(img_mask, Image):
            raise GraphicsError("\n\nGraphicsError: img mask argument for composite must be another goopylib_b Image "
                                f"Object, not {img_mask}")

        self.img_PIL = Img.composite(self.img_PIL, img.img_PIL, img_mask.img_PIL)
        self.transforming_img = self.img_PIL
        self.update_required = True

        self._update_layer()
        return self

    def convert_greyscale(self):
        self.img_PIL = self.img_PIL.convert("LA")
        self.transforming_img = self.img_PIL
        self.update_required = True

        self._update_layer()
        return self

    def convert_binary(self):
        self.img_PIL = self.img_PIL.convert("1")
        self.transforming_img = self.img_PIL
        self.update_required = True

        self._update_layer()
        return self

    # ----------------------------------
    # Flipping Functions

    def flip(self, x_axis=True, y_axis=True):
        if x_axis:
            self.flip_x()
        if y_axis:
            self.flip_y()

        self._update_layer()
        return self

    def flip_x(self):
        self.img_PIL = self.img_PIL.transpose(Img.FLIP_LEFT_RIGHT)
        self.update_required = True

        self._update_layer()
        return self

    def flip_y(self):
        self.img_PIL = self.img_PIL.transpose(Img.FLIP_TOP_BOTTOM)
        self.update_required = True

        self._update_layer()
        return self

    def flip_xy(self):
        self.flip_x()
        self.flip_y()
        return self

    def transverse(self):
        self.img_PIL = self.img_PIL.transpose(Img.TRANSVERSE)
        self.update_required = True

        self._update_layer()
        return self

    def transpose(self):
        self.img_PIL = self.img_PIL.transpose(Img.TRANSPOSE)
        self.update_required = True

        self._update_layer()
        return self

    # ----------------------------------
    # Skew Transformation Functions

    def skew_x(self, scale=0.3, sampling=None, align="center"):
        if sampling is None:
            sampling = Image.default_sampling
        if self.x_skew != scale:
            if sampling not in IMAGE_INTERPOLATIONS.keys():
                raise GraphicsError("\n\nGraphicsError: The resizing sampling argument must be one of "
                                    f"{IMAGE_INTERPOLATIONS}, not {sampling}")
            if not (isinstance(scale, int) or isinstance(scale, float)):
                raise GraphicsError(f"\n\nThe skew scale must be an integer or float, not {scale}")

            if align != "center":
                valid = False
                if align == "left" or align == "right":
                    self.move_to_x(self.anchor[0], align=align)
                else:
                    raise GraphicsError(f"\n\nGraphicsError: Skew align must be one of ['center', 'left', 'right'], "
                                        f"not {align}")

            self.x_skew = scale

            # Code (modified) from https://stackoverflow.com/questions/14177744/how-does-perspective-transformation-work-in-pil
            width, height = self.transforming_img.width, self.transforming_img.height
            new_width = round((1 + abs(scale)) * width)
            size_factor = width / height
            self.img_PIL = self.transforming_img.transform((new_width, height), Img.AFFINE, (1, scale * size_factor,
                                                  (abs(scale) * -width) if scale > 0 else 0, 0, 1, 0),
                                                  IMAGE_INTERPOLATIONS[sampling])
            self.update_required = True

        self._update_layer()
        return self

    def skew_y(self, scale=0.3, sampling=None, align="center"):
        if sampling is None:
            sampling = Image.default_sampling
        if self.y_skew != scale:
            if sampling not in IMAGE_INTERPOLATIONS.keys():
                raise GraphicsError("\n\nGraphicsError: The resizing sampling argument must be one of "
                                    f"{IMAGE_INTERPOLATIONS}, not {sampling}")
            if not (isinstance(scale, int) or isinstance(scale, float)):
                raise GraphicsError(f"\n\nThe skew scale must be an integer or float, not {scale}")

            if align != "center":
                valid = False
                if align == "top" or align == "bottom":
                    self.move_to_y(self.anchor[1], align=align)
                else:
                    raise GraphicsError(f"\n\)nGraphicsError: Skew align must be one of ['center', 'top', 'bottom'], "
                                        f"not {align}")

            self.y_skew = scale

            width, height = self.transforming_img.width, self.transforming_img.height
            new_height = round(height * (abs(scale) + 1))
            size_factor = height / width
            self.img_PIL = self.transforming_img.transform((width, new_height), Img.AFFINE,
                                                  (1, 0, 0, scale * size_factor, 1,
                                                  (abs(scale) * -height) if scale > 0 else 0),
                                                  IMAGE_INTERPOLATIONS[sampling])

            self.update_required = True

        self._update_layer()
        return self

    def skew_xy(self, x_scale=0.3, y_scale=None, sampling=None, x_align="center", y_align=None):
        if y_align is None:
            y_align = x_align
        if y_scale is None:
            y_scale = x_scale

        if sampling is None:
            sampling = Image.default_sampling

        if sampling not in IMAGE_INTERPOLATIONS.keys():
            raise GraphicsError("\n\nGraphicsError: The resizing sampling argument must be one of "
                                f"{IMAGE_INTERPOLATIONS}, not {sampling}")
        if not (isinstance(x_scale, int) or isinstance(x_scale, float)):
            raise GraphicsError(f"\n\nThe skew x_scale must be an integer or float, not {x_scale}")
        if not (isinstance(y_scale, int) or isinstance(y_scale, float)):
            raise GraphicsError(f"\n\nThe skew y_scale must be an integer or float, not {y_scale}")

        if y_align != "center":
            valid = False
            if y_align == "top" or y_align == "bottom":
                self.move_to_y(self.anchor[1], align=y_align)
            else:
                raise GraphicsError(f"\n\nGraphicsError: Skew Y Align must be one of ['center', 'top', 'bottom'], "
                                    f"not {y_align}")
        if x_align != "center":
            valid = False
            if x_align == "left" or x_align == "right":
                self.move_to_x(self.anchor[0], align=x_align)
            else:
                raise GraphicsError(f"\n\nGraphicsError: Skew X Align must be one of ['center', 'left', 'right'], "
                                    f"not {x_align}")

        self.x_skew = x_scale
        self.y_skew = y_scale

        width, height = self.transforming_img.width, self.transforming_img.height
        new_height = round(height * (y_scale + 1))
        new_width = round(width * (x_scale + 1))

        size_factor_h = height / new_width
        size_factor_w = width / new_height
        self.img_PIL = self.transforming_img.transform((new_width, new_height), Img.AFFINE, (
            1, x_scale * size_factor_w, (abs(x_scale) * -width) if x_scale > 0 else 0,
            y_scale * size_factor_h, 1, (abs(y_scale) * -height) if y_scale > 0 else 0),
                                                       IMAGE_INTERPOLATIONS[sampling])

        self.update_required = True

        self._update_layer()
        return self

    # ----------------------------------
    # Resizing Functions

    def resize(self, width, height, sampling=None, _external_call=True):
        if not (isinstance(width, int) or isinstance(width, float)):
            raise GraphicsError(f"\n\nGraphicsError: Resize Width must be an integer or float, not {width}")
        if not (isinstance(height, int) or isinstance(height, float)):
            raise GraphicsError(f"\n\nThe window's height must be an integer or float, not {height}")

        if sampling is None:
            sampling = Image.default_sampling

        if sampling not in IMAGE_INTERPOLATIONS.keys():
            raise GraphicsError("\n\nGraphicsError: The resizing sampling argument must be one of "
                                f"{IMAGE_INTERPOLATIONS}, not {sampling}")

        self.img_PIL = self.img_PIL.resize((ceil(width), ceil(height)), IMAGE_INTERPOLATIONS[sampling])
        self.original_img = self.original_img.resize((ceil(width), ceil(height)), IMAGE_INTERPOLATIONS[sampling])

        if _external_call:
            self.initial_width = width
            self.initial_height = height
        self.update_required = True

        self._update_layer()

        return self

    def resize_height(self, height, preserve_aspect_ratio=False, sampling=None, _external_call=True):

        if not isinstance(preserve_aspect_ratio, bool):
            raise GraphicsError("\n\nGraphicsError: Resizing Preserve Aspect Ratio argument must be a boolean, not "
                                f"{preserve_aspect_ratio}")

        width = height * (self.get_width() / self.get_height()) if preserve_aspect_ratio else self.get_width()
        self.resize(width, height, sampling=sampling, _external_call=_external_call)
        return self

    def resize_width(self, width, preserve_aspect_ratio=False, sampling=None, _external_call=True):

        if not isinstance(preserve_aspect_ratio, bool):
            raise GraphicsError("\n\nGraphicsError: Resizing Preserve Aspect Ratio argument must be a boolean, not "
                                f"{preserve_aspect_ratio}")

        height = width * (self.get_height() / self.get_width()) if preserve_aspect_ratio else self.get_height()
        self.resize(width, height, sampling=sampling, _external_call=_external_call)
        return self

    def resize_factor(self, factor, sampling=None, _external_call=True):
        self.resize_width_factor(factor, sampling, _external_call)
        self.resize_height_factor(factor, sampling, _external_call)
        return self

    def resize_width_factor(self, factor, sampling=None, _external_call=True):
        self.resize(int(self.get_width() * factor), self.get_height(), sampling, _external_call)
        return self

    def resize_height_factor(self, factor, sampling=None, _external_call=True):
        self.resize(self.get_width(), int(self.get_height() * factor), sampling, _external_call)
        return self

    def resize_to_fit(self, obj, sampling=None, _external_call=True):
        self.resize(obj.get_width(), obj.get_height(), sampling=sampling, _external_call=_external_call)
        return self

    def resize_to_fit_width(self, obj, preserve_aspect_ratio=False, sampling=None, _external_call=True):
        self.resize_width(obj.get_width(), preserve_aspect_ratio=preserve_aspect_ratio, sampling=sampling,
                          _external_call=_external_call)
        return self

    def resize_to_fit_height(self, obj, preserve_aspect_ratio=False, sampling=None, _external_call=True):
        self.resize_height(obj.get_height(), preserve_aspect_ratio=preserve_aspect_ratio,
                           sampling=sampling, _external_call=_external_call)
        return self

    # ----------------------------------
    # Blurring & Sharpening Functions

    def blur(self):
        self.img_PIL = self.original_img.filter(filter=ImageFilter.BLUR)
        self.update_required = True

        self._update_layer()
        return self

    def blur_box(self, radius=3):

        if not (isinstance(radius, int) or isinstance(radius, float)):
            raise GraphicsError(f"\n\nGraphicsError: Radius for Blur must be an integer or float, not {radius}")

        self.img_PIL = self.original_img.filter(filter=ImageFilter.BoxBlur(radius))
        self.update_required = True

        self._update_layer()
        return self

    def blur_gaussian(self, radius=3):

        if not (isinstance(radius, int) or isinstance(radius, float)):
            raise GraphicsError(f"\n\nGraphicsError: Radius for Blur must be an integer or float, not {radius}")

        self.img_PIL = self.original_img.filter(filter=ImageFilter.GaussianBlur(radius))
        self.blur_amount = radius
        self.update_required = True

        self._update_layer()
        return self

    def sharpen(self, radius=3, percent=150):

        if not (isinstance(radius, int) or isinstance(radius, float)):
            raise GraphicsError(f"\n\nGraphicsError: Radius for Blur must be an integer or float, not {radius}")
        if not isinstance(percent, int):
            raise GraphicsError(f"\n\nGraphicsError: Percent Blur must be an integer or float, not {percent}")

        self.img_PIL = self.img_PIL.filter(filter=ImageFilter.UnsharpMask(radius=radius, percent=percent))
        self.update_required = True

        self._update_layer()
        return self

    # ----------------------------------
    # Filter Functions

    def filter_contour(self):
        self.img_PIL = self.img_PIL.filter(filter=ImageFilter.CONTOUR)
        self.update_required = True

        self._update_layer()
        return self

    def filter_detail(self):
        self.img_PIL = self.img_PIL.filter(filter=ImageFilter.DETAIL)
        self.update_required = True

        self._update_layer()
        return self

    def filter_emboss(self):
        self.img_PIL = self.img_PIL.filter(filter=ImageFilter.EMBOSS)
        self.update_required = True

        self._update_layer()
        return self

    def filter_find_edges(self):
        self.img_PIL = self.img_PIL.filter(filter=ImageFilter.FIND_EDGES)
        self.update_required = True

        self._update_layer()
        return self

    def filter_sharpen(self):
        self.img_PIL = self.img_PIL.filter(filter=ImageFilter.SHARPEN)
        self.update_required = True

        self._update_layer()
        return self

    def filter_smooth(self):
        self.img_PIL = self.img_PIL.filter(filter=ImageFilter.SMOOTH)
        self.update_required = True

        self._update_layer()
        return self

    def filter_more_smooth(self):
        self.img_PIL = self.img_PIL.filter(filter=ImageFilter.SMOOTH_MORE)
        self.update_required = True

        self._update_layer()
        return self

    def filter_enhance_edge(self):
        self.img_PIL = self.img_PIL.filter(filter=ImageFilter.EDGE_ENHANCE)
        self.update_required = True

        self._update_layer()
        return self

    def filter_more_enhance_edge(self):
        self.img_PIL = self.img_PIL.filter(filter=ImageFilter.EDGE_ENHANCE_MORE)
        self.update_required = True

        self._update_layer()
        return self

    # -------------------------------------------------------------------------
    # GETTER FUNCTIONS

    def get_anchor(self):
        return self.anchor.copy()

    def get_width(self):
        """Returns the width of the image in pixels"""
        return self.img_PIL.width

    def get_height(self):
        """Returns the height of the image in pixels"""
        return self.img_PIL.height

    def get_aspect_ratio(self):
        divisor = gcd(self.get_width(), self.get_height())
        return self.get_width() / divisor, self.get_height() / divisor

    def get_pixel(self, x, y):
        """Returns a list [r,g,b] with the RGB colour values for pixel (x,y)
        r,g,b are in range(256)

        """
        return self.img_PIL.getpixel((x, y))

    def get_format(self):
        return self.img_PIL.format

    def get_mode(self):
        return self.img_PIL.mode

    def get_contrast(self):
        return self.contrast

    # -------------------------------------------------------------------------
    # SETTER FUNCTIONS

    def set_pixel(self, x, y, colour):
        """Sets pixel (x,y) to the given colour

        """
        self.original_img = self.original_img.putpixel((x, y), colour)
        self.img_PIL = self.img_PIL.putpixel((x, y), colour)
        self.update_required = True

        self._update_layer()
        return self

    @staticmethod
    def set_default_sampling(sampling):
        if sampling not in IMAGE_INTERPOLATIONS:
            raise GraphicsError(f"\n\nGraphicsError: sampling to set as default must be one of {IMAGE_INTERPOLATIONS}, "
                                f"not {sampling}")
        Image.default_sampling = sampling
    
    @staticmethod
    def set_texture_path(path):
        Image.texture_path = path



from goopylib.objects.GraphicsObject import GraphicsObject
from goopylib._internal_classes import VectorEquation
from goopylib.styles import *
from goopylib.constants import ARROWS, CAPSTYLES, JOINSTYLES, DASHES

from math import sin, cos, radians, atan

# The Line class is used for drawing straight lines segments between 2 or more points
class Line(GraphicsObject):

    def __init__(self, *p, style=None, outline=None, outline_width=None, arrow=None, capstyle=None, joinstyle=None,
                 cursor="arrow", arrow_shape=None, arrow_scale=None, dash=None, bounds_width=None, layer=0, tag=None):

        self.points = list(p)  # The list of points that define the line segment
        for p in self.points:
            if not isinstance(p, list):  # Checking if all the points specified are indeed list objects
                raise GraphicsError("\n\nGraphicsError: Points for Line class must be lists in the form [x, y], "
                                    f"not {p}")

        if len(p) < 2:  # Making sure there are at least 2 points
            raise GraphicsError(f"\n\nGraphicsError: There must be at least two points to create the line")

        self.low_x, self.low_y = None, None  # These variables are all defined in the get_size() function called below
        self.high_x, self.high_y = None, None  # They store the extreme x & y points farthest from the line's anchor
        self.outline_width, self.height = None, None

        self.get_size()  # Calculates the Width & Height of the line

        self.anchor = None  # The anchor is the center of the line
        self.get_anchor()  # It is calculated as the averages of all the points in the line

        GraphicsObject.__init__(self, options=["joinstyle", "arrowshape", "dash"],
                                cursor=cursor, layer=layer, tag=tag)

        self.arrow_scale_coeff = 1 if arrow_scale is None else arrow_scale

        if arrow is None:
            self.arrow = STYLES["default"]["arrow"]

        elif isinstance(arrow, str):
            arrow = arrow.lower()
            if arrow in ARROWS:  # Check if the Arrow specified is valid
                self.arrow = arrow
            else:
                raise GraphicsError(f"\n\nGraphicsError: capstyle for Line must be one of {ARROWS}, not {arrow}")
        else:
            raise GraphicsError(f"\n\nGraphicsError: arrow option for Line must be one of {ARROWS}, not {arrow}")

        if outline is None:
            self.outline = STYLES["default"]["line outline"]
        elif isinstance(outline, Colour):  # Checking if the option is a colour
            self.outline = outline
        else:  # If not, raise an error
            raise GraphicsError(f"\n\nGraphicsError: The line outline must be a Colour object , not {outline}")

        if outline_width is None:
            self.outline_width = STYLES["default"]["width"]
        elif isinstance(outline_width, int):  # Checking if the option is an integer
            self.outline_width = outline_width
        else:  # If not, raise an error
            raise GraphicsError(f"\n\nGraphicsError: The line width must be an integer, not {outline_width}")
        self.arrow_scale = self.outline_width * self.arrow_scale_coeff

        if capstyle is None:
            self.capstyle = STYLES["default"]["capstyle"]
        elif isinstance(capstyle, str):
            capstyle = capstyle.lower()
            if capstyle in CAPSTYLES:
                self.capstyle = capstyle
        else:
            raise GraphicsError(f"\n\nGraphicsError: capstyle for line must be one of {CAPSTYLES}, not {style}")
        
        if joinstyle is None:
            self.joinstyle = STYLES["default"]["joinstyle"]
        elif isinstance(joinstyle, str):
            joinstyle = joinstyle.lower()
            if joinstyle in JOINSTYLES:
                self.joinstyle = joinstyle
        else:
            raise GraphicsError(f"\n\nGraphicsError: joinstyle for line must be one of {JOINSTYLES}, not {joinstyle}")
        
        if arrow_shape is None:
            self.arrow_shape = STYLES["default"]["arrowshape"]
        elif isinstance(arrow_shape, list):  # Check if the Arrow arrow_shape specified is valid
            if arrow_scale:
                arrow_shape = arrow_shape
                for i in range(len(arrow_shape)):
                    arrow_shape[i] = arrow_shape[i] * self.arrow_scale  # Scaling the arrow to match the width of the line
            self.arrow_shape = arrow_shape  # Changing the arrow
        else:
            raise GraphicsError(f"\n\nGraphicsError: arrow shape for line must be a list, not {arrow_shape}")

        if dash is None:  # Checking if the value is none, if it is, it becomes a solid line
            self.dash = None
        else:
            if dash in DASHES:
                self.dash = DASHES[dash]

            elif isinstance(dash, list):
                for num in dash:
                    if 0 < num < 266 and isinstance(num, int):
                        continue
                    else:
                        raise GraphicsError(f"\n\nGraphicsError: custom dash numbers must be integers within 1 & 255")
                self.dash = dash
            else:
                raise GraphicsError(f"\n\nLine dash must be a tuple or a string referencing a dash "
                                    f"(one of {DASHES.keys()}), not {dash}")

        self.bounds_width = None
        self.set_bounds_width(bounds_width)

        self.segments = None  # These 2 variables are defined in the _update() method
        self.equations = None
        self._update()

        # This makes sure that if the user uses the set_fill() function on the line, it references set_outline()
        self.set_fill = self.set_outline

    def __repr__(self):
        return f"Line({self.points})"

    def __dir__(self):
        return "see https://github.com/BhavyeMathur/goopylib/wiki/Lines-&-Curved-Lines!"

    def _draw(self, canvas, options):
        # Converting all the coordinates to Window coordinates to account for stretching, changed coords, etc.
        points = [canvas.to_screen(point[0], point[1]) for point in self.points]
        return canvas.create_line(points, arrow=self.arrow, fill=self.outline, width=self.outline_width,
                                  capstyle=self.capstyle, joinstyle=self.joinstyle, arrowshape=self.arrow_shape,
                                  dash=self.dash)  # Creating the line!

    def _move(self, dx, dy):
        for i in range(len(self.points)):  # Going through all the points in the line
            self.points[i] += [dx, dy]  # And then moving them
        self.anchor = self.get_anchor()  # Recalculating the center of the line

    def _rotate(self, dr, scaling="bicubic", center=None):
        if center is None:
            center = self.get_anchor()  # Getting the center of rotation

        for point in self.points:
            # Formula from (https://stackoverflow.com/questions/2259476/rotating-a-point-about-another-point-2d)
            p = point.clone()
            point[0] = cos(radians(dr)) * (p[0] - center[0]) - sin(radians(dr)) * (p[1] - center[1]) + center[0]
            point[1] = sin(radians(dr)) * (p[0] - center[0]) + cos(radians(dr)) * (p[1] - center[1]) + center[1]

    def _update(self):  # Called by the GraphicsObject class whenever the object rotates/moves and the __init__ function
        # These store the points for the line segments present in the line.
        # As this line object can have more than 2 points, it can be divided into segments of 2 points
        self.segments = [(self.points[i], self.points[i + 1]) for i in range(len(self.points) - 1)]
        self.equations = []  # The variable to store the Vector Equations of the line segments
        for line in self.segments:
            # For more information about these steps, see https://www.desmos.com/calculator/esn97hdwbe
            try:
                slope = (line[0][1] - line[1][1]) / (line[0][0] - line[1][0])  # Finding the slope of the line segment
                shift = - line[0][1] / slope  # Finding the x-shift

                # The height of the line to get it to be the correct amount of wide when rotated
                width = self.bounds_width / (cos(atan(slope)))

                smaller = min([line[0][0], line[1][0]])  # Which point has a smaller/larger x value?
                larger = max([line[0][0], line[1][0]])

                self.equations.append(VectorEquation(  # Creating the Vector Equation of this line segment
                    f"({slope} * (x - {shift} - {line[0][0]}) + {width}/2 > y > "
                    f"{slope} * (x - {shift} - {line[0][0]}) - {width}/2) and ({smaller} < x < {larger})"))

            except GraphicsError:  # The Line is Horizontal

                smaller = min([line[0][1], line[1][1]])  # Which point has a smaller/larger y value?
                larger = max([line[0][1], line[1][1]])

                self.equations.append(VectorEquation(
                    f"({line[0][0] + self.bounds_width/2} > x > {line[1][0] - self.bounds_width/2}) and"
                    f"({larger} > y > {smaller})"))

            except ZeroDivisionError:  # The Line is Vertical

                smaller = min([line[0][0], line[1][0]])  # Which point has a smaller/larger x value?
                larger = max([line[0][0], line[1][0]])

                self.equations.append(VectorEquation(
                    f"({line[0][1] + self.bounds_width/2}> y > {line[0][1] - self.bounds_width/2}) "
                    f"and ({smaller} < x < {larger})"))

        self.get_anchor()

    # -------------------------------------------------------------------------
    # GETTER FUNCTIONS

    def get_anchor(self):
        # To calculate the center, it simply calculates the Average of all the Points
        self.anchor = [0, 0]  # Starts the anchor at (0, 0)
        for point in self.points:
            self.anchor[0] += point[0]  # Sums up all the Point values
            self.anchor[1] += point[1]

        self.anchor[0] //= len(self.points)  # Divides by number of Points
        self.anchor[1] //= len(self.points)

        return self.anchor

    def get_width(self):
        # To calculate the width, it finds the smallest & largest x points and gets the width
        self.low_x = self.points[0][0]  # Setting the High & Low point to the first point
        self.high_x = self.low_x

        for point in self.points[1:]:  # Going through all the other points
            if point[0] < self.low_x:  # Checking if this point is smaller
                self.low_x = point[0]
            elif point[0] > self.high_x:  # Checking if this point is larger
                self.high_x = point[0]
        self.outline_width = abs(self.high_x - self.low_x)  # Getting the width

        return self.outline_width

    def get_height(self):
        # To calculate the height, it finds the smallest & largest y points and gets the height
        self.low_y = self.points[0][1]  # Setting the High & Low point to the first point
        self.high_y = self.low_y

        for point in self.points[1:]:  # Going through all the other points
            if point[1] < self.low_y:  # Checking if this point is smaller
                self.low_y = point[1]
            elif point[1] > self.high_y:  # Checking if this point is larger
                self.high_y = point[1]
        self.height = abs(self.high_y - self.low_y)  # Getting the height

        return self.height

    # These next functions return the value from the Object's config
    def get_outline_width(self):
        return self.outline_width

    def get_fill(self):
        return self.outline

    def get_arrow(self):
        return self.arrow

    def get_outline(self):
        return self.outline

    def get_capstyle(self):
        return self.capstyle

    def get_joinstyle(self):
        return self.joinstyle

    def get_arrow_shape(self):
        return self.arrow_shape

    def get_bounds_width(self):
        return self.bounds_width

    # -------------------------------------------------------------------------
    # SETTER FUNCTIONS

    def set_arrow(self, option):
        if option.lower() in ARROWS:  # Check if the Arrow specified is valid
            self.arrow = option
        else:
            raise GraphicsError(f"\n\nGraphicsError: arrow for line must be one of {ARROWS}, not {option}")

    # These arrow functions reference the set_arrow() function to set the arrow
    def remove_arrows(self):
        self.set_arrow("none")

    def set_arrow_both(self):
        self.set_arrow("both")

    def set_arrow_first(self):
        self.set_arrow("first")

    def set_arrow_last(self):
        self.set_arrow("last")

    def set_outline(self, outline):
        if isinstance(outline, Colour):  # Checking if the option is a colour
            self.outline = outline
        else:  # If not, raise an error
            raise GraphicsError(f"\n\nGraphicsError: The line outline must be a Colour object , not {outline}")

    def set_outline_width(self, outline_width):
        if isinstance(outline_width, int):  # Checking if the option is an integer
            self.outline_width = outline_width
        else:  # If not, raise an error
            raise GraphicsError(f"\n\nGraphicsError: The line width must be an integer, not {outline_width}")

        self.arrow_scale = self.outline_width * self.arrow_scale_coeff
            
    def set_capstyle(self, style):
        style = style.lower()
        if style in CAPSTYLES:  # Check if the Arrow specified is valid
            self.capstyle = style  # Changing the arrow
        else:
            raise GraphicsError(f"\n\nGraphicsError: capstyle for line must be one of {CAPSTYLES}, not {style}")
    
    def set_joinstyle(self, joinstyle):
        joinstyle = joinstyle.lower()
        if joinstyle in JOINSTYLES:  # Check if the Arrow specified is valid
            self.joinstyle = joinstyle  # Changing the arrow
        else:
            raise GraphicsError(f"\n\nGraphicsError: joinstyle for line must be one of {JOINSTYLES}, not {joinstyle}")
            
    def set_arrow_shape(self, shape, scale=True):
        if isinstance(scale, bool):
            if isinstance(shape, list):  # Check if the Arrow shape specified is valid
                if scale:
                    shape = shape
                    for i in range(len(shape)):
                        shape[i] = shape[i] * self.arrow_scale  # Scaling the arrow to match the width of the line
                self.arrowshape = shape  # Changing the arrow
            else:
                raise GraphicsError(f"\n\nGraphicsError: arrow shape for line must be a list, not {shape}")
        else:
            raise GraphicsError(f"\n\nGraphicsError: scale argument must be a boolean, not {scale}")

    def set_dash(self, dash):
        if dash is None:  # Checking if the value is none, if it is, it becomes a solid line
            self.dash = None
        else:
            if dash in DASHES:
                self.dash = DASHES[dash]

            elif isinstance(dash, list):
                for num in dash:
                    if 0 < num < 266 and isinstance(num, int):
                        continue
                    else:
                        raise GraphicsError(f"\n\nGraphicsError: custom dash numbers must be integers within 1 & 255")
                self.dash = dash
            else:
                raise GraphicsError(f"\n\nLine dash must be a tuple or a string referencing a dash "
                                    f"(one of {DASHES.keys()}), not {dash}")

    def set_bounds_width(self, width):
        if width is None:
            self.bounds_width = self.get_outline_width()
        else:
            if not (isinstance(width, float) or isinstance(width, int)):
                raise GraphicsError("\n\nGraphicsError: The bounds width for the line must be an integer or float, "
                                    f"not {width}")
            self.bounds_width = width

    # -------------------------------------------------------------------------
    # OTHER FUNCTIONS

    def is_clicked(self, mouse_pos):
        # Go through every line segment equation  and check if that segment has been clicked
        for eq in self.equations:
            if eq.is_clicked(mouse_pos):
                return True
        return False



from goopylib.objects.GraphicsObject import GraphicsObject
from goopylib.styles import STYLES, global_style
from goopylib.colours import Colour
from goopylib.util import GraphicsError

from tkinter import Text as tkText
from tkinter import WORD as tkWORD
from tkinter import END as tkEND
from tkinter import Frame as tkFrame

class MultilineEntry(GraphicsObject):
    def __init__(self, p, line_length, style=None, fill=None, outline_width=None, relief=None, cursor="arrow",
                 font_size=None, font_style=None, font_colour=None, number_of_lines=2, align="center", tag=None,
                 bounds=None, select_colour=None, font_face=None, text=""):

        self.widget = None
        self.text = text

        if style is None:
            self.style = global_style
        else:
            self.style = style

        if isinstance(fill, Colour):
            self.fill = fill
        elif fill in STYLES[self.style].keys():
            self.fill = STYLES[self.style][fill]
        else:
            if "fill" in STYLES[self.style].keys():
                self.fill = STYLES[self.style]["fill"]
            else:
                self.fill = STYLES["default"]["fill"]

        if isinstance(select_colour, Colour):
            self.select_colour = select_colour
        elif select_colour in STYLES[self.style].keys():
            self.select_colour = STYLES[self.style][select_colour]
        else:
            if "select colour" in STYLES[self.style].keys():
                self.select_colour = STYLES[self.style]["select colour"]
            else:
                self.select_colour = STYLES["default"]["select colour"]

        if isinstance(font_colour, Colour):
            self.font_colour = font_colour
        elif font_colour in STYLES[self.style].keys():
            self.font_colour = STYLES[self.style][font_colour]
        else:
            if "font colour" in STYLES[self.style].keys():
                self.font_colour = STYLES[self.style]["font colour"]
            else:
                self.font_colour = STYLES["default"]["font colour"]

        if isinstance(outline_width, int):
            self.outline_width = outline_width
        elif outline_width in STYLES[self.style].keys():
            self.outline_width = STYLES[self.style][outline_width]
        else:
            if "width" in STYLES[self.style].keys():
                self.outline_width = STYLES[self.style]["entry width"]
            else:
                self.outline_width = STYLES["default"]["entry width"]

        if font_style in STYLES[self.style].keys():
            self.font_style = STYLES[self.style][font_style]
        elif isinstance(font_style, str):
            self.font_style = font_style
        else:
            if "font style" in STYLES[self.style].keys():
                self.font_style = STYLES[self.style]["font style"]
            else:
                self.font_style = STYLES["default"]["font style"]

        if font_face in STYLES[self.style].keys():
            self.font = STYLES[self.style][font_face]
        elif isinstance(font_face, str):
            self.font = font_face
        else:
            if "font face" in STYLES[self.style].keys():
                self.font = STYLES[self.style]["font face"]
            else:
                self.font = STYLES["default"]["font face"]

        self.relief = relief
        self.cursor = cursor

        self.font_size = 0
        self.initial_font_size = 0
        self.set_size(font_size, True)

        self.align = align
        self.line_length = line_length

        self.number_of_lines = number_of_lines

        self.anchor = p.clone()

        super().__init__((), tag=tag, bounds=bounds)

    def _draw(self, canvas, options):
        x, y = self.anchor

        frm = tkFrame(canvas.master)

        self.set_size(self.initial_font_size / canvas.trans.x_scale, False)

        self.widget = tkText(frm, bg=self.fill, bd=self.outline_width, relief=self.relief, width=self.line_length,
                             cursor=self.cursor, exportselection=0, fg=self.font_colour, wrap=tkWORD,
                             font=(self.font, self.font_size, self.font_style), height=self.number_of_lines,
                             insertbackground=self.font_colour)
        self.widget.pack()
        self.set_text(self.text)

        x, y = canvas.to_screen(x, y)

        return canvas.create_window(x, y, window=frm)

    def get_anchor(self):
        return self.anchor

    def set_size(self, font_size=None, _set_font_size=False):
        if isinstance(font_size, int):
            self.font_size = font_size
        elif font_size in STYLES[self.style].keys():
            self.font_size = STYLES[self.style][font_size]
        else:
            if "font size" in STYLES[self.style].keys():
                self.font_size = STYLES[self.style]["font size"]
            else:
                self.font_size = STYLES["default"]["font size"]

        if _set_font_size:
            self.initial_font_size = round(self.font_size)

        return self

    def get_font_size(self):
        return self.font_size

    def get_font_face(self):
        return self.font

    def get_font_colour(self):
        return self.font_colour

    def get_font_style(self):
        return self.font_style

    def get_selection_colour(self):
        return self.select_colour

    def get_outline_width(self):
        return self.outline_width

    def get_fill(self):
        return self.fill

    def get_text(self):
        if self.drawn and self.graphwin.is_open():
            return self.widget.get("1.0", "end-1c")
        else:
            raise GraphicsError("\n\nGraphicsError: get_text() function for the Multiline Entry object can only be "
                                "used if the object is drawn and the window is open")

    def set_fill(self, colour):
        self.fill = colour
        if self.widget:
            self.widget.config(bg=colour)
        self._update_layer()
        return self

    def set_face(self, font_face):
        if font_face in STYLES[self.style].keys():
            self.font = STYLES[self.style][font_face]
        elif isinstance(font_face, str):
            self.font = font_face
        else:
            if "font face" in STYLES[self.style].keys():
                self.font = STYLES[self.style]["font face"]
            else:
                self.font = STYLES["default"]["font face"]

        if self.widget:
            self.widget.config(font=(self.font, self.font_size, self.font_style))
        self._update_layer()
        return self

    def set_font_size(self, font_size, set_initial_font_size=True):
        if isinstance(font_size, int):
            self.font_size = font_size
        elif font_size in STYLES[self.style].keys():
            self.font_size = STYLES[self.style][font_size]
        else:
            if "font size" in STYLES[self.style].keys():
                self.font_size = STYLES[self.style]["font size"]
            else:
                self.font_size = STYLES["default"]["font size"]

        if set_initial_font_size:
            self.initial_font_size = round(font_size)

        if self.widget:
            self.widget.config(font=(self.font, self.font_size, self.font_style))
        self._update_layer()
        return self

    def set_font_style(self, font_style):
        if font_style in STYLES[self.style].keys():
            self.font_style = STYLES[self.style][font_style]
        elif isinstance(font_style, str):
            self.font_style = font_style
        else:
            if "font style" in STYLES[self.style].keys():
                self.font_style = STYLES[self.style]["font style"]
            else:
                self.font_style = STYLES["default"]["font style"]

        if self.widget:
            self.widget.config(font=(self.font, self.font_size, self.font_style))
        self._update_layer()
        return self

    def set_text_colour(self, colour):
        self.font_colour = colour
        if self.widget:
            self.widget.config(fg=colour)
        self._update_layer()
        return self

    def set_text(self, text):
        self.text = str(text)
        if self.widget:
            self.widget.delete(1.0, tkEND)
            self.widget.insert(tkEND, self.text)
        return self



from setuptools import setup, Extension
from distutils.core import setup, Extension

import os
import platform
import shutil


def create_release():
    with open("README.md", "r") as fh:
        long_description = fh.read()

    setup(
        name='goopylib',
        package_dir={':math': 'goopylib/math', ":objects": 'goopylib/objects', ":applications": 'goopylib/applications',
                     ":sound": 'goopylib/sound', ":physics": 'goopylib/physics'},

        packages={"goopylib": "goopylib",
                  "goopylib/math": "goopylib/math",
                  "goopylib/objects": "goopylib/objects",
                  "goopylib/applications": "goopylib/applications",
                  "goopylib/sound": "goopylib/sound",
                  "goopylib/physics": "goopylib/physics"},

        include_package_data=True,

        version='1.1.277a23',
        license='MIT License',

        description='A simple-yet-powerful 2D graphics framework built on top of Tkinter capable of creating '
                    'good-looking, modern GUIs, games, and animations.',
        long_description=long_description,
        long_description_content_type="text/markdown",

        author='Bhavye Mathur',
        author_email='bhavyemathur@gmail.com',
        url='https://github.com/BhavyeMathur/goopylib',
        download_url='https://github.com/BhavyeMathur/goopylib/archive/v1.1.277-alpha.tar.gz',
        keywords=['Tkinter', '2D Graphics', 'Python GUI', 'Game Creator', 'Graphics Library'],
        install_requires=['pillow'],
        project_urls={"Bug Tracker": "https://github.com/BhavyeMathur/goopylib/issues",
                      "Documentation": "https://github.com/BhavyeMathur/goopylib/wiki",
                      "Source Code": "https://github.com/BhavyeMathur/goopylib/tree/master/goopylib"},
        classifiers=[
            'Intended Audience :: Developers',
            'Topic :: Software Development :: Build Tools',
            'License :: OSI Approved :: MIT License',
        ],
        python_requires=">=3.6"
    )


def setup_extension(name, sources, output):
    name = "c_" + name
    setup(name=name, ext_modules=[Extension(name, sources=sources)])

    system = platform.system()

    try:
        if system == "Windows":
            try:
                os.remove(f"goopylib/{output}{name}.pyd")
            except FileNotFoundError:
                pass
            os.rename(f"build/lib.win-amd64-3.8/{name}.cp38-win_amd64.pyd", f"goopylib/{output}{name}.pyd")

        elif system == "Darwin":
            try:
                os.remove(f"goopylib/{output}{name}.o")
            except FileNotFoundError:
                pass
            os.rename(f"build/lib.macosx-10.9-x86_64-3.8/{name}.cpython-38-darwin.so", f"goopylib/{output}{name}.so")

            shutil.rmtree("dist")
            shutil.rmtree(f"{name}.egg-info")

    except FileNotFoundError as e:
        print(e)

    shutil.rmtree("build")


#setup_extension("colours", ["goopylib/colours.c"], "")
# setup_extension("easing", ["goopylib/math/Easing.c"], "math/")
# setup_extension("bezier_curve", ["goopylib/math/bezier_curve.c"], "math/")

create_release()
# To check release: twine check dist/*
# To upload release: twine upload dist/*

# To create release: python setup.py sdist bdist_wheel
# To build Extension: python setup.py build

# Create text wrapping



from goopylib.objects.GraphicsObject import GraphicsObject
from goopylib.util import GraphicsError

from math import cos, sin


class CycleButton(GraphicsObject):
    def __init__(self, *states, state=0, disabled_graphic=None, autoflush=True, layer=None, tag=None, bounds=None):
        self.states = list(states)
        self.state = state

        self.is_object_disabled = False
        self.disabled_graphic = disabled_graphic
        self.drawn_graphic = None

        self.autoflush = autoflush

        self.graphic = states[state]

        self.number_of_states = len(self.states)

        number_of_states = 0
        self.anchor = [0, 0]
        for state in self.states:
            state_anchor = state.get_anchor()
            if state_anchor is not None:
                self.anchor[0] += state_anchor[0]
                self.anchor[1] += state_anchor[1]
                number_of_states += 1

        if number_of_states != 0:
            self.anchor = [self.anchor[0] // number_of_states, self.anchor[1] // number_of_states]

        GraphicsObject.__init__(self, (), tag=tag, bounds=bounds)
        GraphicsObject.cyclebutton_instances.add(self)

        if layer is not None:
            self.set_layer(layer)

    def __repr__(self):
        return f"CycleButton({self.graphic} and {len(self.states)} other states)"

    def __iter__(self):
        for state in self.states:
            yield state

    def __getitem__(self, item):
        return self.states[item]

    def __len__(self):
        return self.number_of_states

    def _draw(self, canvas=None, options=None):
        self.states[self.state].draw(canvas, _internal_call=True)
        self.drawn_graphic = self.states[self.state]

        for graphic in self.states:
            graphic.graphwin = canvas
        if self.disabled_graphic is not None:
            self.disabled_graphic.graphwin = canvas

        return self

    def _undraw(self, set_blinking=True):
        self.drawn_graphic.undraw(set_blinking=set_blinking)
        return self

    def base_undraw(self):
        self.drawn_graphic.undraw()

    def destroy(self):
        GraphicsObject.objects.discard(self)
        GraphicsObject.object_layers[self.layer].discard(self)
        GraphicsObject.draggable_objects.discard(self)
        GraphicsObject.cursor_objects.discard(self)

        GraphicsObject.cyclebutton_instances.discard(self)

        if self in GraphicsObject.redraw_on_frame[self.layer]:
            GraphicsObject.redraw_on_frame[self.layer].remove(self)

        for state in self.states:
            state.destroy()

        self.drawn = False
        self.graphwin = None
        self.id = None

        return self

    def _rotate(self, dr, sampling="bicubic", center=None):
        for graphic in self.states:
            graphic.rotate(dr, sampling=sampling, center=center)
        if self.disabled_graphic is not None:
            self.disabled_graphic.rotate(dr, sampling=sampling, center=center)

    def _move(self, dx, dy):
        self.anchor[0] += dx
        self.anchor[1] += dy
        for graphic in self.states:
            graphic.move(dx, dy)
        if self.disabled_graphic is not None:
            self.disabled_graphic.move(dx, dy)

    def click(self):
        if not self.is_object_disabled:
            self.state += 1
            self.state %= len(self.states)

            self.graphic = self.states[self.state]

            self._update_layer()
        return self

    def is_clicked(self, mouse_pos):
        if self.bounds is None:
            return self.graphic.is_clicked(mouse_pos)
        else:
            return self.bounds.is_clicked(mouse_pos)

    def change_graphic(self, element, value):
        self.states[element] = value
        self._update_layer()

    # -------------------------------------------------------------------------
    # LAYERING SYSTEM FUNCTIONS

    def move_up_layer(self, layers=1):
        if not isinstance(layers, int):
            raise GraphicsError(f"\n\nGraphicsError: layers to move up must be an integer, not {layers}")
        if layers < 0:
            raise GraphicsError("\n\nGraphicsError: layers to move up must be greater than (or equal to) 0, "
                                f"not {layers}")

        GraphicsObject.object_layers[self.layer].remove(self)
        self.layer += layers

        while self.layer > len(GraphicsObject.object_layers) - 1:
            GraphicsObject.object_layers.append([])
        GraphicsObject.object_layers[self.layer].add(self)

        for obj in self.states:
            obj.move_up_layer(layers=layers)
        if self.disabled_graphic is not None:
            self.disabled_graphic.move_up_layer(layers=layers)
        return self

    def move_down_layer(self, layers=1):
        if not isinstance(layers, int):
            raise GraphicsError(f"\n\nGraphicsError: layers to move down must be an integer, not {layers}")
        if layers < 0:
            raise GraphicsError("\n\nGraphicsError: layers to move down must be greater than (or equal to) 0, "
                                f"not {layers}")

        GraphicsObject.object_layers[self.layer].remove(self)
        self.layer += layers
        if self.layer < 0:
            self.layer = 0

        while self.layer > len(GraphicsObject.object_layers) - 1:
            GraphicsObject.object_layers.append([])
        GraphicsObject.object_layers[self.layer].add(self)

        for obj in self.states:
            obj.move_down_layer(layers=layers)
        if self.disabled_graphic is not None:
            self.disabled_graphic.move_down_layer(layers=layers)
        return self

    def set_layer(self, layer=0):
        if not isinstance(layer, int):
            raise GraphicsError(f"\n\nGraphicsError: layer to set to must be an integer, not {layer}")
        if layer < 0:
            raise GraphicsError("\n\nGraphicsError: layer to set to must be greater than (or equal to) 0, "
                                f"not {layer}")

        for obj in self.states:
            obj.set_layer(layer=layer)
        if self.disabled_graphic is not None:
            self.disabled_graphic.set_layer(layer=layer)

        GraphicsObject.object_layers[self.layer].remove(self)
        while layer > len(GraphicsObject.object_layers) - 1:
            GraphicsObject.object_layers.append({*()})
        GraphicsObject.object_layers[layer].add(self)

        self.layer = layer

        return self

    # -------------------------------------------------------------------------
    # ENABLING & DISABLING FUNCTIONS

    def enable(self):
        self.is_object_disabled = False

        self.graphic = self.states[self.state]

        self.rotation = self.graphic.rotation
        self.cosrotation = cos(self.rotation / 57.2958)
        self.sinrotation = sin(self.rotation / 57.2958)
        self._update_layer()
        return self

    def disable(self):
        self.is_object_disabled = True

        if self.disabled_graphic is not None:
            self.graphic = self.disabled_graphic

            self.rotation = self.graphic.rotation
            self.cosrotation = cos(self.rotation / 57.2958)
            self.sinrotation = sin(self.rotation / 57.2958)
            self._update_layer()
        return self

    def toggle_enabled(self):
        self.is_object_disabled = not self.is_object_disabled
        if self.is_object_disabled:
            self.disable()
        else:
            self.enable()
        return self

    def set_enabled(self, enable=True):
        if enable:
            self.enable()
        else:
            self.disable()
        return self

    def is_enabled(self):
        return not self.is_object_disabled

    def is_disabled(self):
        return self.is_object_disabled

    # -------------------------------------------------------------------------
    # SETTER FUNCTIONS

    def set_state(self, state):
        if state != self.state:
            if self.graphic in GraphicsObject.redraw_on_frame[self.layer]:
                GraphicsObject.redraw_on_frame[self.layer].remove(self.graphic)

            self.state = state
            self.graphic = self.states[state]

            self.rotation = self.graphic.rotation
            self.cosrotation = cos(self.rotation / 57.2958)
            self.sinrotation = sin(self.rotation / 57.2958)

            self._update_layer()

        return self

    def add_state(self, state):
        self.states.append(state)
        self.number_of_states += 1
        number_of_states = 0
        self.anchor = [0, 0]
        for state in self.states:
            state_anchor = state.get_anchor()
            if state_anchor is not None:
                self.anchor[0] += state_anchor[0]
                self.anchor[1] += state_anchor[1]
                number_of_states += 1

        self.anchor = [self.anchor[0] // number_of_states, self.anchor[1] // number_of_states]
        return self

    def remove_state(self, state):
        self.number_of_states -= 1
        number_of_states = 0
        self.anchor = [0, 0]
        for state in self.states:
            state_anchor = state.get_anchor()
            if state_anchor is not None:
                self.anchor[0] += state_anchor[0]
                self.anchor[1] += state_anchor[1]
                number_of_states += 1

        self.anchor = [self.anchor[0] // number_of_states, self.anchor[1] // number_of_states]

        self.states.remove(state)
        return self

    def pop_state(self, index):
        self.number_of_states -= 1
        number_of_states = 0
        self.anchor = [0, 0]
        for state in self.states:
            state_anchor = state.get_anchor()
            if state_anchor is not None:
                self.anchor[0] += state_anchor[0]
                self.anchor[1] += state_anchor[1]
                number_of_states += 1

        self.anchor = [self.anchor[0] // number_of_states, self.anchor[1] // number_of_states]
        return self.states.pop(index)

    def set_object(self, obj):
        if obj not in self.states:
            if obj in GraphicsObject.tagged_objects:
                obj = GraphicsObject.tagged_objects[obj]
            else:
                raise GraphicsError(f"\n\nGraphicsError: The object you have specified ({obj}) is not a valid state "
                                    f"for this CycleButton")

        if obj != self.graphic:
            if self.graphic in GraphicsObject.redraw_on_frame[self.layer]:
                GraphicsObject.redraw_on_frame[self.layer].remove(self.graphic)

            self.state = self.states.index(obj)
            self.graphic = obj

            self.rotation = self.graphic.rotation
            self.cosrotation = cos(self.rotation / 57.2958)
            self.sinrotation = sin(self.rotation / 57.2958)

            self._update_layer()

        return self

    def get_object(self):
        return self.states[self.state]

    # -------------------------------------------------------------------------
    # GETTER FUNCTIONS

    def get_anchor(self):
        return self.graphic.get_anchor()

    def get_state(self):
        return self.state

    def get_graphics(self):
        return self.states

    def set_contrast(self, level):
        for state in self.states:
            state.set_contrast(level)
        self._update_layer()
        return self

    def change_contrast(self, level):
        for state in self.states:
            state.change_contrast(level)
        self._update_layer()
        return self

    def reset_contrast(self):
        self.set_contrast(0)
        self._update_layer()

    # -------------------------------------------------------------------------
    # IMAGE MANIPULATION FUNCTIONS

    def crop(self, left=0, top=0, right=500, down=500, align="center"):
        for state in self.states:
            state.crop(left=left, top=top, right=right, down=down, align=align)
        self._update_layer()
        return self

    # Blending & Compositing Functions

    def blend(self, state, alpha):
        for state in self.states:
            state.blend(state, alpha)
        self._update_layer()
        return self

    def alpha_composite(self, other):
        for state in self.states:
            state.alpha_composite(other)
        self._update_layer()
        return self

    def composite(self, other, state_mask):
        for state in self.states:
            state.composite(other, state_mask)
        self._update_layer()
        return self

    def convert_greyscale(self):
        for state in self.states:
            state.convert_greyscale()
        self._update_layer()
        return self

    def convert_binary(self):
        for state in self.states:
            state.convert_binary()
        self._update_layer()
        return self

    # ----------------------------------
    # Flipping Functions

    def flip(self, x_axis=True, y_axis=True):
        for state in self.states:
            state.flip(x_axis=x_axis, y_axis=y_axis)
        self._update_layer()
        return self

    def flip_x(self):
        for state in self.states:
            state.flip_x()
        self._update_layer()
        return self

    def flip_y(self):
        for state in self.states:
            state.flip_y()
        self._update_layer()
        return self

    def flip_xy(self):
        for state in self.states:
            state.flip_xy()
        self._update_layer()
        return self

    def transverse(self):
        for state in self.states:
            state.transverse()
        self._update_layer()
        return self

    def transpose(self):
        for state in self.states:
            state.transpose()
        self._update_layer()
        return self

    # ----------------------------------
    # Skew Transformation Functions

    def skew_x(self, scale=0.3, sampling="bicubic", align="center"):
        for state in self.states:
            state.skew_x(scale=scale, sampling=sampling, align=align)
        self._update_layer()
        return self

    def skew_y(self, scale=0.3, sampling=None, align="center"):
        for state in self.states:
            state.skew_y(scale=scale, sampling=sampling, align=align)
        self._update_layer()
        return self

    def skew_xy(self, x_scale=0.3, y_scale=None, sampling=None, x_align="center", y_align=None):
        for state in self.states:
            state.skew_xy(x_scale=x_scale, y_scale=y_scale, sampling=sampling, x_align=x_align, y_align=y_align)
        self._update_layer()
        return self

    # ----------------------------------
    # Resizing Functions

    def resize(self, width, height, sampling=None, _external_call=True):
        for state in self.states:
            state.resize(width, height, sampling=sampling, _external_call=_external_call)
        self._update_layer()
        return self

    def resize_height(self, height, preserve_aspect_ratio=False, sampling=None, _external_call=True):
        for state in self.states:
            state.resize_height(height, preserve_aspect_ratio=preserve_aspect_ratio, sampling=sampling,
                                _external_call=_external_call)
        self._update_layer()
        return self

    def resize_width(self, width, preserve_aspect_ratio=False, sampling=None, _external_call=True):
        for state in self.states:
            state.resize_width(width, preserve_aspect_ratio=preserve_aspect_ratio, sampling=sampling,
                               _external_call=_external_call)
        self._update_layer()
        return self

    def resize_factor(self, factor, sampling=None, _external_call=True):
        for state in self.states:
            state.resize_factor(factor, sampling=sampling, _external_call=_external_call)
        self._update_layer()
        return self

    def resize_width_factor(self, factor, sampling=None, _external_call=True):
        for state in self.states:
            state.resize_width_factor(factor, sampling=sampling, _external_call=_external_call)
        self._update_layer()
        return self

    def resize_height_factor(self, factor, sampling=None, _external_call=True):
        for state in self.states:
            state.resize_width_factor(factor, sampling=sampling, _external_call=_external_call)
        self._update_layer()
        return self

    def resize_to_fit(self, obj, sampling=None, _external_call=True):
        for state in self.states:
            state.resize_to_fit(obj, sampling=sampling, _external_call=_external_call)
        self._update_layer()
        return self

    def resize_to_fit_width(self, obj, preserve_aspect_ratio=False, sampling=None, _external_call=True):
        for state in self.states:
            state.resize_to_fit_width(obj, preserve_aspect_ratio=preserve_aspect_ratio, sampling=sampling,
                                      _external_call=_external_call)
        self._update_layer()
        return self

    def resize_to_fit_height(self, obj, preserve_aspect_ratio=False, sampling=None, _external_call=True):
        for state in self.states:
            state.resize_to_fit_height(obj, preserve_aspect_ratio=preserve_aspect_ratio, sampling=sampling,
                                       _external_call=_external_call)
        self._update_layer()
        return self

    # ----------------------------------
    # Blurring & Sharpening Functions

    def blur(self):
        for state in self.states:
            state.blur()
        self._update_layer()
        return self

    def blur_box(self, radius=3):
        for state in self.states:
            state.blur_box(radius=radius)
        self._update_layer()
        return self

    def blur_gaussian(self, radius=3):
        for state in self.states:
            state.blur_gaussian(radius=radius)
        self._update_layer()
        return self

    def sharpen(self, radius=3, percent=150):
        for state in self.states:
            state.sharpen(radius=radius, percent=percent)
        self._update_layer()
        return self

    # ----------------------------------
    # Filter Functions

    def filter_contour(self):
        for state in self.states:
            state.filter_contour()
        self._update_layer()
        return self

    def filter_detail(self):
        for state in self.states:
            state.filter_detail()
        self._update_layer()
        return self

    def filter_emboss(self):
        for state in self.states:
            state.filter_emboss()
        self._update_layer()
        return self

    def filter_find_edges(self):
        for state in self.states:
            state.filter_find_edges()
        self._update_layer()
        return self

    def filter_sharpen(self):
        for state in self.states:
            state.filter_sharpen()
        self._update_layer()
        return self

    def filter_smooth(self):
        for state in self.states:
            state.filter_smooth()
        self._update_layer()
        return self

    def filter_more_smooth(self):
        for state in self.states:
            state.filter_more_smooth()
        self._update_layer()
        return self

    def filter_enhance_edge(self):
        for state in self.states:
            state.filter_enhance_edge()
        self._update_layer()
        return self

    def filter_more_enhance_edge(self):
        for state in self.states:
            state.more_enhance_edge()
        self._update_layer()
        return self



from tkinter import Tk as tkTk
from time import time as timetime

from PIL.Image import LANCZOS, BOX, BICUBIC, BILINEAR, NEAREST, HAMMING

import platform

# platform.system() gives Linux for Linux, Windows for Windows, and Darwin for MacOS
# https://stackoverflow.com/questions/1854/what-os-am-i-running-on
OS = platform.system()

# For the list of cursors visit:
CURSORS_RAW = {"x_cursor", "arrow", "based_arrow_down", "based_arrow_up", "boat", "bogosity", "bottom_left_corner",
               "bottom_right_corner", "bottom_side", "bottom_tee", "box_spiral", "center_ptr", "circle", "clock",
               "coffee_mug", "cross", "cross_reverse", "crosshair", "diamond_cross", "dot", "dotbox", "double_arrow",
               "draft_large", "draft_small", "draped_box", "exchange", "fleur", "gobbler", "gumby", "hand1", "hand2",
               "heart", "icon", "iron_cross", "left_ptr", "left_side", "left_tee", "leftbutton", "ll_angle", "lr_angle",
               "man", "middlebutton", "mouse", "pencil", "pirate", "plus", "question_arrow", "right_ptr", "right_side",
               "right_tee", "rightbutton", "rtl_logo", "sailboat", "sb_down_arrow", "sb_h_double_arrow",
               "sb_left_arrow", "sb_right_arrow", "sb_up_arrow", "sb_v_double_arrow", "shuttle", "sizing", "spider",
               "spraycan", "star", "target", "tcross", "top_left_arrow", "top_left_corner", "top_right_corner",
               "top_side", "top_tee", "trek", "ul_angle", "umbrella", "ur_angle", "watch", "xterm"}

CURSORS_WINDOWS_RAW = {"no", "starting", "size", "size_ne_sw", "size_ns", "size_nw_se", "size_we", "uparrow", "wait"}
CURSORS_MACOSX_RAW = {"copyarrow", "aliasarrow", "contextualmenuarrow", "text", "cross-hair", "closedhand", "openhand",
                      "pointinghand", "resizeleft", "resizeright", "resizeleftright", "resizeup", "resizedown",
                      "resizeupdown", "none", "notallowed", "poof", "countinguphand", "countingdownhand",
                      "countingupanddownhand", "spinning"}

CURSORS = {"x": "x_cursor", "arrow": "arrow", "based-down-arrow": "based_arrow_down",
           "based-up-arrow": "based_arrow_up", "boat": "boat", "maze_symbol": "bogosity",
           "bottom-left-corner-arrow": "bottom_left_corner", "bottom-right-corner-arrow": "bottom_right_corner",
           "download": "bottom_side", "bottom-t": "bottom_tee", "spiral": "box_spiral", "up-arrow-thin": "center_ptr",
           "donut": "circle", "clock": "clock", "mug": "coffee_mug", "double-plus": "cross",
           "inverted-plus": "cross_reverse", "crosshair": "crosshair", "diamond-plus": "diamond_cross", "circle": "dot",
           "box": "dotbox", "double-arrow": "double_arrow", "flipped-arrow": "draft_large",
           "thin-flipped-arrow": "draft_small", "diamond-box": "draped_box", "rotating-arrows": "exchange",
           "move": "fleur", "bird": "gobbler", "stickman": "gumby", "pointer": "hand1", "select": "hand2",
           "heart": "heart", "blackbox": "icon", "iron-cross": "iron_cross", "small-arrow": "left_ptr",
           "left-download": "left_side", "left-t": "left_tee", "left-button": "leftbutton",
           "bottom-left-corner": "ll_angle", "bottom-right-corner": "lr_angle", "volcano": "man",
           "middle-button": "middlebutton", "mouse": "mouse", "pencil": "pencil", "pirate": "pirate", "plus": "plus",
           "question-arrow": "question_arrow", "pointer-arrow": "right_ptr", "right-download": "right_side",
           "right-t": "right_tee", "right-button": "rightbutton", "maze": "rtl_logo", "sailboat": "sailboat",
           "down-arrow": "sb_down_arrow", "horizontal-resizing-arrow": "sb_h_double_arrow",
           "left-arrow": "sb_left_arrow", "right-arrow": "sb_right_arrow", "up-arrow": "sb_up_arrow",
           "vertical-resizing-arrow": "sb_v_double_arrow", "shuttle": "shuttle", "sizing": "sizing", "spider": "spider",
           "spraycan": "spraycan", "star": "star", "target": "target", "t-cross": "tcross",
           "flipped-pointer": "top_left_arrow", "top-left-corner-arrow": "top_left_corner",
           "top-right-corner-arrow": "top_right_corner", "top-download": "top_side", "top-t": "top_tee",
           "virus": "trek",
           "top-left-corner": "ul_angle", "umbrella": "umbrella", "top-right-corner": "ur_angle", "loading": "watch",
           "text-cursor": "xterm"}
CURSORS_WINDOWS = {"cancel": "no"}
CURSORS_MACOSX = {}

RELIEF = {"flat", "sunken", "raised", "groove", "ridge"}
ARROWS = {"both", "none", "first", "last"}

CAPSTYLES = {"round", "butt", "projecting"}
JOINSTYLES = {"bevel", "round", "miter"}
DASHES = {"solid": None, "dash": (5, ), "long dash": (8, ), "dot": (1, 3), "dashdot": (5, 1, 1, 1),
          "long dashdot": (8, 1, 1, 1)}

DUPLICATES_METRICS = {"Colour Animation": {"Change", "Start", "Time", "Update", "Initial", "Easing", "Final"},
                      "2D Animation": {"Change", "Start", "Time", "Update", "Initial", "EasingX", "EasingY", "Final"},
                      "1D Animation": {"Change", "Start", "Time", "Update", "Initial", "Easing", "Final"}}

CURVE_INTERPOLATIONS = {"linear", "cosine", "cubic", "hermite", "spline"}
IMAGE_INTERPOLATIONS = {"bicubic": BICUBIC, "bilinear": BILINEAR, "box": BOX, "nearest": NEAREST, "hamming": HAMMING,
                        "lanczos": LANCZOS}

ARC_TYPES = {"arc", "pieslice", "chord"}

ALIGN_OPTIONS = {"top", "left", "right", "bottom", "center", "topleft", "topright", "bottomleft", "bottomright"}

INTERPOLATIONS = {"linear", "poly", "poly_in", "poly_out", "sin", "sin_in", "sin_out", "circle", "circle_in",
                  "circle_out", "back", "back_in", "back_out", "bounce", "bounce_in", "bounce_out", "elastic",
                  "elastic_in", "elastic_out", "exponential", "exponential_in", "exponential_out"}

ALL_CHARACTERS = {'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j',
                  'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', 'A', 'B', 'C', 'D',
                  'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X',
                  'Y', 'Z', '!', '"', '#', '$', '%', '&', '\\', "'", '(', ')', '*', '+', ',', '-', '.', '/', ':', ';',
                  '<', '=', '>', '?', '@', '{', '\'', '"', '}', '^', '_', '`', '{', '|', '}', '~', ':', " "}

##########################################################################
# global variables and functions

_root = tkTk()
_root.withdraw()

_update_lasttime = timetime()
